/* $Id: dmpico.c,v 1.12 1998/04/13 20:31:57 queinnec Exp $
 * ///////////////////////////////////////////////////////////////////////
 *                    //          //
 *                   // DMeroon  //
 *                  //          //
 * A continuation-based causal distributed class-based object system.
 * Christian.Queinnec@polytechnique.fr
 * ///////////////////////////////////////////////////////////////////////
 */

/* DOCUMENTATION 
 * This is the Pico runtime interface with DMeroon.
 * It uses the generic dmbind.{c,h} files.
 */

#include "dmeroon.h"

DM_DefineStringWithCstring(pico_version,
  "$Id: dmpico.c,v 1.12 1998/04/13 20:31:57 queinnec Exp $");

/* RESOURCE DM_allocate_bare_instance
 * This function allocates an instance of a given class with a given size.
 * This size is expressed in bytes and only corresponds to the size of the
 * content of the DMeroon object. To this size is added the size of 
 * the DMeroon header (two void* fields) and the size required by the 
 * support language (held in the DM_PrefixSize() macro).

 SIGNATURE:
 (Object (DM_allocate_bare_instance Class nat3 Object))

 */

DM_SupportPointer 
DM_allocate_bare_instance (DM_SupportPointer wrapped_class, 
                           DM_nat3           size,
                           DM_SupportPointer wrapped_reference ) {
  DM_Class *class     = DM_FromSupport(wrapped_class);
  void     *wrapped_o;
  int      sz         = DM_OffsetRealignToSomeAlignment(size, 2);
  
  if ( DM_HasOption(class, DM_CLASS_IS_STRINGENT) ) {
    sz += DM_ROOM_FOR_REALIGNMENT;
    wrapped_o = (void*)pico_allocate(DM_PrefixSize() + sz);
    if ( wrapped_o == NULL )
      return DM_Abort(context, DME_ALLOCATION_PROBLEM);
    wrapped_o = (DM_SupportPointer)
      DM_OffsetRealign((DM_SupportAddress)wrapped_o);
  } else {
    wrapped_o = (void*)pico_allocate( DM_PrefixSize() + sz );
    if ( wrapped_o == NULL )
      return DM_Abort(context, DME_ALLOCATION_PROBLEM);
  }

  /* Structure the object for Pico */
  ((pico_struct*) wrapped_o)->_header = DM_HEADER;
  ((pico_struct*) wrapped_o)->_size   = sz;

 /* Initialize DMeroon header, its class and proxy, etc. */
  return DM_structure_bare_instance(PicoInsertTag(wrapped_o,2),
                                    wrapped_class, 
                                    size,
                                    wrapped_reference );
}

/* RESOURCE DM_print_char_star
 * This procedure allows DMeroon to use the printing machinery of the 
 * support language. string is the address of a C string, len is the number 
 * characters to be printed (if zero the whole string (until the next null
 * character) is printed), wrapped_port is an XXX value telling where to 
 * print (for debug, this may be stdout or stderr). The function returns
 * true if it succeeds printing.

 SIGNATURE:
 (boolean (DM_print_char_star string nat3 anything))

 */

DM_SupportBoolean
DM_print_char_star (char              *string, 
                    DM_nat3           len,
                    DM_SupportPointer wrapped_port) 
{
  int n = len ? len : strlen(string);

  if ( wrapped_port == NULL ) return;

  DM_DebugAtLevel(0,
     if (    (wrapped_port == (DM_SupportPointer)stdout)
          || (wrapped_port == (DM_SupportPointer)stderr) ) {
       FILE *stream = (FILE*) wrapped_port;
       
       fprintf(stream, "%*s", n, string);
       fflush(stream);
       return DM_TRUE;
     }
                  );

  if ( pico_output_portp( wrapped_port ) == DM_TRUE ) {
    FILE *fout = ((pico_port*)(PicoRemoveTag(wrapped_port, 2)))->file;
    fputs(string, fout);
    return DM_TRUE;
  }

  { DM_Connection *connection = DM_FromSupport(wrapped_port);
  if ( DM_IsA(connection, Connection) ) {
    DM_OutBuffer *emission  = connection->emission;
      
    DM_print_char_star_in_outbuffer(string, len, DM_ToSupport(emission));
  }
  }
  
  return DM_Abort(context, DME_UNPRINTABLE_OBJECT);
}

/* RESOURCE DM_raise_exception
 * The error function. It takes an error code, a message identifying the 
 * problem as well as the filename and the linenumber indicating where
 * the error was signalled. This function may be very crude and may exit
 * the current process (after reporting the error).

 SIGNATURE:
 (anything (DM_raise_exception nat3 string string nat3))

 */

DM_SupportPointer
DM_raise_exception (DM_nat3 code, 
                    char    *message, 
                    char    *file, 
                    DM_nat3 line) {
  C_scheme_error(1);

  DM_DebugAtLevel(10,
    fflush(stdout);
    fprintf(stderr,"Error %lu in %s(%lu)\n%s\n",code,file,line,message);
);
  exit(1);
}

/* RESOURCE DM_initialize
 * This is the single function that initializes entirely DMeroon from the
 * support language. DM_initialize initializes individually the various
 * parts of DMeroon. It returns the Site instance describing reflectively
 * the DMeroon site. It also checks some assumptions to ease porting, you
 * may remove these tests when successful.

 SIGNATURE:
 (Site (DM_initialize))

 */

DM_SupportPointer
DM_initialize () {

  /* Only initialize once! */
  static DM_nat1 already_done = 0;
  if ( already_done ) return DM_ToSupport(DM_ReferenceTo(DM_site));
  already_done = 1;
  
  /* Initialize now the various packages (in order).  */
  DM_bind_initialize();
  DM_types_initialize();  
  DM_kernel_initialize();
  DM_extr_initialize();
  DM_net_initialize();
  DM_site_initialize();
  DM_comm_initialize();

  /* Return the site object. */
  return DM_ToSupport(DM_ReferenceTo(DM_site));
}

/* RESOURCEE PICO_allocate_double
 * Doubles are not yet present!
 */

PICO_OBJ
PICO_allocate_double(double d) {
  return DM_TRUE;
}

void
DM_set_exit_finalization (DM_SupportPointer wrapped_exit) {
}


/* end of dmpico.c */
