;;; $Id: sig2pico.scm,v 1.7 1995/12/15 19:22:00 queinnec Exp $
;;;///////////////////////////////////////////////////////////////////////
;;;                      //          //
;;;                     // DMeroon  //
;;;                    //          //
;;; A continuation-based causal distributed class-based object system.
;;; Christian.Queinnec@polytechnique.fr
;;;///////////////////////////////////////////////////////////////////////

;;; This file converts signatures into appropriate interfaces for
;;; foreign languages such as module for Pico.

;;;{{{ Mapping types for Pico

(set! types-mapping
  '(;; Remove this case since many languages have problems with that 
    ;; return type:   (void         . void)  !!!
    ;; Dont confuse string and String!
    (string      . str)
    (char        . char)
    (fixnum      . int)
    (floatnum    . PICO_OBJ)
    (anything    . PICO_OBJ)
    (boolean     . PICO_OBJ)
    (nat1        . int)
    (nat2        . int)
    (nat3        . int)
    (int1        . int)
    (int2        . int)
    (int3        . int)
    ) )

;;;}}}
;;;{{{ Main function

;;; Read signatures, produce two C file to be compiled as an extension
;;; of Pico. The first will define new symbols bound to C functions
;;; or C variables, the second introduces C functions (checking
;;; arguments) wrapping C macros or C functions.

(define (prepare-c-4pico sigfile cfile scmfile)
  (complete-types-mapping)
  (let ((signatures (read-signatures-file sigfile)))
    ;; Declare the C macros for the interpreter.
    (call-with-output-file cfile
      (lambda (out)
        (produce-c-comment "Automatically generated (Don't edit!)" out)
        (display "#include \"dmeroon.h\"" out)
        (newline out)
        (for-each (lambda (sig)
                    (produce-c-comment sig out)
                    (produce-c-definition sig out)
                    (newline out) )
                  signatures )
        (produce-c-eof out) ) )
    ;; Wrap C functions and variables within Scheme definitions.
    (call-with-output-file scmfile
      (lambda (out)
        (produce-scheme-comment "Automatically generated (Don't edit!)" out)
        (for-each (lambda (sig)
                    (produce-scheme-comment sig out)
                    (let ((e (produce-body sig)))
                      (if e (begin 
                              (write e out)
                              (newline out) )) ) )
                  signatures )
        (produce-scheme-eof out) ) )
    'done ) )

;;;}}}
;;;{{{ Convert C macros into C functions.
;;; Code inspired from sig2osc.scm

(define (produce-c-definition sig out)
  (if (pair? (cadr sig))
      (if (equal? 'macro (car (cadr sig)))
          (produce-c-macro sig out)
          (produce-c-function sig out) )
      (produce-c-constant sig out) ) )

(define (produce-c-function sig out)
  #f )

(define (produce-c-constant sig out)
  #f )

(define (produce-c-macro sig out)
  (produce-regular-function-converter
   `(,(car sig) ,(cdr (cadr sig)))
   out ) )

(define (produce-regular-function-converter sig out)
  (let* ((name      (car (cadr sig)))
         (types     (cdr (cadr sig)))
         (variables (convert-sigtypes-to-names types)) )
    (display (convert-type (car sig)) out)
    (display " Pico_" out)
    (display name out)
    (display-list-with-individual-prefix "PICO_OBJ " ", " variables out)
    (display " {" out)
    (newline out)
    ;; Proceed to the real treatment now.
    (produce-return-code (car sig) name types variables out)
    ;; finish the definition
    (display ";" out)
    (newline out)
    (display "}" out) ) )

;;; Produce the return expression, calls the function, decodes its
;;; arguments and encodes the result.

(define (produce-return-code return-type name types variables out)
  (display "  return " out)
  ;; convert the result to Pico
  (if (DMeroon-class-type? return-type)
      (display "(" out)
      (case return-type
        ((nat1 nat2 nat3 int1 int2 int3 netnat2 netnat3
          translation displacement options index offset 
          fixnum )
         (display "DM_ToSupportFixnum(" out) )
        ((char)
         (display "DM_ToSupportCharacter(" out) )
        ((floatnum)
         (display "DM_ToSupportFloat(" out) )
        ((boolean anything)
         (display "(" out) )
        ;; otherwise no coercion on the returned value
        (else (sigtools-error "Cannot coerce back" return-type)) ) )
  ;; output the function (or macro) name
  (display name out)
  (display "(" out)
  ;; convert arguments from Pico now:
  (let loop ((types types)
             (variables variables) )
    (if (pair? types)
        (begin 
          (produce-argument (car types) (car variables) out)
          (if (pair? (cdr types))
              (begin (display ", " out)
                     (loop (cdr types) (cdr variables)) ) ) ) ) )              
  (display "))" out) )

(define (produce-argument type variable out)
  (if (char-upper-case? (string-ref (symbol->string type) 0))
      (display variable out)
      (case type
        ((nat1 nat2 nat3 int1 int2 int3 netnat2 netnat3
          translation displacement options index offset 
          fixnum )
         (display "DM_FromSupportFixnum(" out)
         (display variable out)
         (display ")" out) )
        ((floatnum)
         (display "DM_FromSupportFloat(" out)
         (display variable out)
         (display ")" out) )
        ((string)
         (display "OsStringBytes(" out)
         (display variable out)
         (display ")" out) )
        ((char)
         (display "DM_FromSupportCharacter(" out)
         (display variable out)
         (display ")" out) )
        ((anything)
         (display variable out) )
        (else (sigtools-error "Cannot decode" type)) ) ) )

;;;}}}
;;;{{{ Build a Scheme function invoking the C function.
;;; Code inspired from sig2bgl.scm

;;; Convert a signature into a function that checks the type of its
;;; arguments. Return #f if there is nothing to emit.

(define (produce-body sig)
  (if (pair? (cadr sig))
      (if (equal? 'macro (car (cadr sig)))
          (produce-macro-body sig)
          (produce-function-body sig) )
      (produce-constant-body sig) ) )

(define (produce-constant-body sig)
  (let* ((type (car sig))
         (name (cadr sig))
         (scheme-name (convert-name name)) )
    ;; scheme-name == PicoInsertTag(&_name_object,2)
    `(define-inlinable-var ,scheme-name 
       (tag-address (asm-address ,(symbol->string (symbol-concatenate
                                                   "_" name "_object" )))
                    2 ) ) ) )

(define (produce-function-body sig)
  (let* ((type      (car sig))
         (call      (cadr sig))
         (name      (car call))
         (types     (cdr call))
         (variables (convert-sigtypes-to-names types)) )
    `(begin
       (define-foreign (,name . ,(map convert-type types))
	 ,(string-append "_" (symbol->string name))
	 ,(convert-type type) )
       (define (,(convert-name name) . ,variables)
         ,(coerce-from 
           type `(,name . ,(map-coerce-to types variables)) ) ) ) ) )

;;; Macros are for internal use, they do not check the type of their 
;;; arguments (to avoid infinite regression).

(define (produce-macro-body sig)
  (let* ((type      (car sig))
         (call      (cdr (cadr sig)))
         (name      (car call))
         (types     (cdr call))
         (variables (convert-sigtypes-to-names types)) )
    `(begin
       ;; no need to do conversion here because they are done in the C file!
       (define-foreign (,name . ,(map (lambda (t) 'PICO_OBJ) types) )
         ,(string-append "_Pico_" (symbol->string name))
         PICO_OBJ ) 
       (define (,(convert-macro-name name) . ,variables)
	 (,name . ,variables) ) ) ) )

;;; Transform the result of a foreign function.
;;; In fact accept it as it is.

(define (coerce-from from exp)
  exp )

;;; Due to the possible dotted argument list, define a special mapper
;;; and take care of the possibly dotted variable. In fact this only
;;; occurs for the signature of DM_allocate in Bigloo.

(define (map-coerce-to types variables)
  (if (pair? types)
      (cons (coerce-to (car types) (car variables))
            (map-coerce-to (cdr types) (cdr variables)) )
      (if (null? types)
          '() 
          (list variables) ) ) )

;;; Check an argument held in variable VAR to be of DMeroon type TYPE.

(define (coerce-to type var)
  (if (DMeroon-class-type? type)
      (if (eq? type 'Object)
          `(if (DM_is_object ,var)
               ,var
               (DM-error ,(string-append "Wrong type: should be" 
                                         (symbol->string type) )
                         ,var ) )
          `(if (and (DM_is_object ,var)
                    (DM_is_a ,var ,(symbol-concatenate "DM-" type "-class")) )
               ,var
               (DM-error ,(string-append "Wrong type: should be" 
                                         (symbol->string type) )
                         ,var ) ) )
      (case type
        ((anything stream void) var)
        ((string) `(if (string? ,var) 
                       ,var 
                       (DM-error "Not a string" ,var) ))
        ((char) `(if (char? ,var) 
                     ,var
                     (DM-error "Not a character" ,var) ))
        ((boolean) `(if (boolean? ,var)
                        ,var
                        (DM-error "Not a boolean" ,var) ))
        ((vector) `(if (vector? ,var)
                       ,var
                       (DM-error "Not a vector" ,var) ))
        ((fixnum) `(if (integer? ,var) 
                       ,var
                       (DM-error "Not a fixnum" ,var) ))
        ((floatnum) `(if (real? ,var) 
                         ,var
                         (DM-error "Not a floatnum" ,var) ))
        ((nat1 index)
         `(if (and (integer? ,var)
                   (>= ,var 0)
                   (<= ,var 255) )
              ,var
              (DM-error "Not a nat1" ,var) ))
        ((nat2 netnat2 translation)
         `(if (and (integer? ,var)
                   (>= ,var 0)
                   (<= ,var 65535) )
              ,var
              (DM-error "Not a nat2" ,var) ))
        ((nat3 netnat3 options offset displacement)
         `(if (and (integer? ,var)
                   (>= ,var 0)
                   ;; This number is read as -1 by OScheme on Irix.     FUTURE
                   ;; Use elk instead of OScheme !
                   (if (< 0 4294967295) (<= ,var 4294967295) #t) )
              ,var
              (DM-error "Not a nat3" ,var) ))
        ((int1)
         `(if (and (integer? ,var)
                   (>= ,var -127)
                   (<= ,var +127) )
              ,var
              (DM-error "Not a int1" ,var) ))
        ((int2)
         `(if (and (integer? ,var)
                   (>= ,var -32767)
                   (<= ,var +32767) )
              ,var
              (DM-error "Not a int2" ,var) ))
        ((int3)
         ;; ATTENTION, bigloo on 32bit word machines, read +/-2147483648
         ;; as 0. In that case avoid the check.
         `(if (and (integer? ,var)
                   (if (< -2147483648 0) (>= ,var -2147483648) #t)
                   (if (< 0 +2147483648) (<= ,var +2147483648) #t) )
              ,var
              (DM-error "Not a int3" ,var) ))
        (else (sigtools-error "Cannot coerce to " type)) ) ) )

;;; end of sig2pico.scm
